const puppeteer = require("puppeteer");
const sleep = require("sleep-promise");
const { clearInterval } = require("timers");
const fs = require("fs");

const HEADLESS = true;
const SCREENSHOT_INTERVAL_MS = 3000;
const USERINFO_INTERVAL_MS = 2000;

// global state variables, because they are great
var stateRoomUrl = null;
var roomPage = null;
var roomScreenShotPath = null;
var screenShotInterval = null;

process.on("message", async ({ eventName, data }) => {
  switch (eventName) {
    case "attachTo":
      await handleAttachToRoom(data);
      break;
    case "sendMessage":
      await handleSendMessage(data);
      break;
    case "detach":
      await handleDetach();
      break;
    default:
      console.warn(
        'Received event "' + eventName + '", but found no way to handle it.'
      );
      break;
  }
});

async function handleAttachToRoom({
  username,
  password,
  displayName,
  roomUrl,
  screenShotPath,
}) {
  stateRoomUrl = roomUrl;
  try {
    // Attach browser to room
    const page = await createFreshPage();
    const authenticatedPage = await signIn(roomUrl, username, password, page);
    const pageInRoom = await joinRoom(authenticatedPage, roomUrl, displayName);
    roomPage = pageInRoom;

    // Set up periodic screenshotting
    screenShotInterval = setInterval(
      handleTakeScreenshot,
      SCREENSHOT_INTERVAL_MS,
      screenShotPath
    );
    roomScreenShotPath = screenShotPath;

    // Set up periodic user info gathering:
    gatherUserInfoInterval = setInterval(
      handleGetRoomInfo,
      USERINFO_INTERVAL_MS
    );

    // Message parent that attachment was successful
    process.send({ eventName: "attachToSuccess", data: null });
  } catch (error) {
    // Crash if attaching to room was not successful
    console.error(error);
    process.send({ eventName: "attachToError", data: error.toString() });
    await sleep(100);
    process.exit(1);
  }
}

async function handleTakeScreenshot(pathToSaveTo) {
  try {
    await screenShotPresentationArea(roomPage, pathToSaveTo);
  } catch (error) {
    console.warn("Unable to take screenshot of room", stateRoomUrl);
  }
}

async function handleGetRoomInfo() {
  try {
    const promises = [
      getUserCount(roomPage),
      getUserList(roomPage),
      getRoomName(roomPage),
    ];
    const userCount = await promises[0];
    const userList = await promises[1];
    const roomName = await promises[2];
    process.send({
      eventName: "roomInfo",
      data: { count: userCount, users: userList, name: roomName },
    });
  } catch (error) {
    console.error(
      "Could not get user info. This might happen, if the meeting get's closed.\n",
      error
    );
  }
}

async function handleSendMessage({ content }) {
  process.send({
    eventName: "debug",
    data: "Received content to send to chat: " + content,
  });
  try {
    await sendChatMessage(roomPage, content);
    process.send({ eventName: "sendMessageSuccess", data: null });
  } catch (error) {
    console.error(error);
    process.send({ eventName: "sendMessageError", data: error });
  }
}

async function handleDetach() {
  clearInterval(screenShotInterval);
  clearInterval(gatherUserInfoInterval);
  await roomPage.close();
  try {
    await fs.promises.unlink(roomScreenShotPath);
  } catch (error) {}
  process.send({ eventName: "detachSuccess", data: null });
  await sleep(100);
  process.exit(0);
}
// ----------------------------
// Utility functions
// ----------------------------

async function createFreshPage() {
  const browser = await puppeteer.launch({
    headless: HEADLESS,
    defaultViewport: {
      width: 1200,
      height: 600,
      isLandscape: true,
    },
  });
  const context = await browser.createIncognitoBrowserContext();
  return context.newPage();
}

async function joinRoom(page, roomUrl, displayName = "Puppet") {
  await page.goto(roomUrl);
  console.log("joinRoom > Gone to start page for", roomUrl);

  // Set displayname for program
  await page.type("input.join-form", displayName);

  // Join and load room
  await page.click("button#room-join");
  //await sleep(3000);

  const dismissButton = await page.waitForSelector(
    'button[aria-describedBy="modalDismissDescription"]'
  );
  console.log("joinRoom > Joined room");
  // Click audio choice panel away
  await dismissButton.click();

  console.log("joinRoom > clicked audio choice away.");
  return page;
}

async function sendChatMessage(page, messageText) {
  const chatInputHandle = await page.$("textarea#message-input");
  const chatSendButtonHandle = await page.$(
    'form > div > button[type="submit"]'
  );

  // Send a text message
  await chatInputHandle.type(messageText);
  await chatSendButtonHandle.click();
  console.log("sendChatMessage > sent Message");
}

async function signIn(bbbUrl, username, password, page = null) {
  // Create page if not done already
  if (page === null) {
    page = await createFreshPage();
  }

  await page.goto(bbbUrl);

  await page.waitForSelector("a.sign-in-button");
  const signInButton = await page.$("a.sign-in-button");
  if (signInButton === null) {
    console.log(
      "SignIn > Can't find sign-in button: User is probably already authenticated."
    );
    return page;
  }

  await page.click("a.sign-in-button");
  await page.waitForNavigation();
  console.log("SignIn > NAVIGATED TO LOGIN PAGE");

  await page.type("input#session_username", username);
  await page.type("input#session_password", password);
  await page.click("input.signin-button");
  console.log("SignIn > entered credentials");

  await page.waitForNavigation();
  console.log("SignIn > Probably logged in.");

  return page;
}

async function screenShotPresentationArea(authenticatedPage, path) {
  const presentationAreaHandle = await authenticatedPage.$("div#container");
  presentationAreaHandle.screenshot({ path, type: "jpeg", quality: 33 });
}

async function getUserCount(authenticatedPage) {
  const selector = 'div[role="complementary"] > div > div > h2';
  const h2Handle = (await authenticatedPage.$$(selector))[2];
  return h2Handle.evaluate(
    (node) => node.innerText.split("(")[1].split(")")[0]
  );
}

async function getUserList(authenticatedPage) {
  const users = [];

  const selector = 'div[class*="userItemContents"]';
  const userHandles = await authenticatedPage.$$(selector);
  for (const userHandle of userHandles) {
    const avatarHandle = await userHandle.$('div[class*="avatar"]');
    const classes = (await avatarHandle.evaluate((node) => node.className))
      .split(" ")
      .map((wholeClass) => wholeClass.split("--")[0])
      .filter((name) => name !== "avatar");
    const userNameHandle = await userHandle.$(
      'span[class*="userNameMain"] > span'
    );
    const userName = await userNameHandle.evaluate((node) => node.innerText);

    users.push({
      classes,
      name: userName,
    });
  }
  return users;
}

async function getRoomName(authenticatedPage) {
  const roomNameHandle = await authenticatedPage.$(
    'h1[class*="presentationTitle"]'
  );
  if (roomNameHandle === null) {
    process.send({ eventName: "sessionClosed", data: null });
    await sleep(100);
    process.exit(0);
  }
  return await roomNameHandle.evaluate((node) => node.innerText);
}
